# paraphase
# Author: Xiao Chen <xchen@pacificbiosciences.com>


import os
import sys
import argparse
import json
import yaml
import logging
import datetime
import shutil
import pysam
import subprocess
import traceback
import random
import multiprocessing as mp
from argparse import RawTextHelpFormatter
from functools import partial
from paraphase.genome_depth import GenomeDepth
from paraphase.prepare_bam_and_vcf import (
    BamRealigner,
    BamTagger,
    VcfGenerater,
    TwoGeneVcfGenerater,
)
import paraphase
from paraphase import genes
from .phaser import Phaser

logging.basicConfig(level=logging.DEBUG)


class Paraphase:
    def __init__(self):
        self.samtools = None
        self.minimap2 = None

    @staticmethod
    def get_version():
        with open(os.path.join(os.path.dirname(__file__), "__init__.py")) as f:
            for line in f:
                if "version" in line:
                    return line.split('"')[1]
        return None

    def parse_configs(self, region_config=None, genome_build="38"):
        """
        Parse config files
        region_config: parameters for each region of interest
        gene_config: specifies some additional analyses triggered with some genes
        """
        data_path = os.path.join(os.path.dirname(__file__), "data", genome_build)
        gene_config = os.path.join(os.path.dirname(data_path), "genes.yaml")
        if region_config is None:
            region_config = os.path.join(data_path, "config.yaml")
        self.region_config_parsed = None
        with open(region_config, "r") as f:
            self.region_config_parsed = yaml.safe_load(f)
        self.accepted_genes = list(self.region_config_parsed.keys())
        with open(gene_config, "r") as f:
            gene_config_parsed = yaml.safe_load(f)
        self.genes_to_call = gene_config_parsed.get("genes_to_call")
        self.no_genome_depth_genes = gene_config_parsed.get("no_genome_depth_genes")
        self.no_vcf_genes = gene_config_parsed.get("no_vcf_genes")
        ## need to update check_sex_genes based on chromosome
        self.check_sex_genes = gene_config_parsed.get("check_sex_genes")
        self.two_reference_regions_genes = gene_config_parsed.get(
            "two_reference_regions_genes"
        )

    def process_gene(
        self,
        gene_list,
        configs,
        outdir,
        tmpdir,
        gdepth,
        bam,
        sample_sex,
        novcf,
        prog_cmd,
        gene1only=False,
    ):
        """Workflow for each region"""
        phaser_calls = {}
        for gene in gene_list:
            try:
                sample_id = bam.split("/")[-1].split(".")[0]
                if gene == "smn1":
                    phaser = genes.Smn1Phaser(
                        sample_id, tmpdir, gdepth, bam, sample_sex
                    )
                elif gene == "rccx":
                    phaser = genes.RccxPhaser(
                        sample_id, tmpdir, gdepth, bam, sample_sex
                    )
                elif gene == "pms2":
                    phaser = genes.Pms2Phaser(
                        sample_id, tmpdir, gdepth, bam, sample_sex
                    )
                elif gene == "strc":
                    phaser = genes.StrcPhaser(
                        sample_id, tmpdir, gdepth, bam, sample_sex
                    )
                elif gene == "ncf1":
                    phaser = genes.Ncf1Phaser(
                        sample_id, tmpdir, gdepth, bam, sample_sex
                    )
                elif gene == "cfc1":
                    phaser = genes.Cfc1Phaser(
                        sample_id, tmpdir, gdepth, bam, sample_sex
                    )
                elif gene == "neb":
                    phaser = genes.NebPhaser(sample_id, tmpdir, gdepth, bam, sample_sex)
                elif gene == "ikbkg":
                    phaser = genes.IkbkgPhaser(
                        sample_id, tmpdir, gdepth, bam, sample_sex
                    )
                elif gene == "f8":
                    phaser = genes.F8Phaser(sample_id, tmpdir, gdepth, bam, sample_sex)
                elif gene == "opn1lw":
                    phaser = genes.Opn1lwPhaser(
                        sample_id, tmpdir, gdepth, bam, sample_sex
                    )
                elif gene == "hba":
                    phaser = genes.HbaPhaser(sample_id, tmpdir, gdepth, bam, sample_sex)
                else:
                    phaser = Phaser(sample_id, tmpdir, gdepth)

                config = configs[gene]
                logging.info(
                    f"Running analysis for {gene} for sample {sample_id} at {datetime.datetime.now()}..."
                )
                logging.info(
                    f"Realigning reads for {gene} for sample {sample_id} at {datetime.datetime.now()}..."
                )
                bam_realigner = BamRealigner(bam, tmpdir, config, prog_cmd)
                bam_realigner.write_realign_bam()

                logging.info(
                    f"Phasing haplotypes for {gene} for sample {sample_id} at {datetime.datetime.now()}..."
                )

                phaser.set_parameter(config)
                phaser_call = phaser.call()._asdict()
                phaser_calls.setdefault(gene, phaser_call)

                logging.info(
                    f"Tagging reads for {gene} for sample {sample_id} at {datetime.datetime.now()}..."
                )
                bam_tagger = BamTagger(sample_id, tmpdir, config, phaser_call)
                bam_tagger.write_bam(random_assign=True)

                # make a tagged bamlet for the second gene region
                if gene in self.two_reference_regions_genes:
                    logging.info(
                        f"Realigning to gene2 region for {gene} for sample {sample_id} at {datetime.datetime.now()}..."
                    )
                    bam_realigner.write_realign_bam(gene2=True)
                    bam_tagger.write_bam(random_assign=True, gene2=True)

                if novcf is False and gene not in self.no_vcf_genes:
                    logging.info(
                        f"Generating VCFs for {gene} for sample {sample_id} at {datetime.datetime.now()}..."
                    )
                    if gene in self.two_reference_regions_genes and gene1only is False:
                        vcf_generater = TwoGeneVcfGenerater(
                            sample_id,
                            outdir,
                            phaser_call,
                        )
                        vcf_generater.set_parameter(
                            config, tmpdir=tmpdir, prog_cmd=prog_cmd
                        )
                        vcf_generater.run()
                    else:
                        vcf_generater = VcfGenerater(
                            sample_id,
                            outdir,
                            phaser_call,
                        )
                        vcf_generater.set_parameter(
                            config, tmpdir=tmpdir, prog_cmd=prog_cmd
                        )
                        vcf_generater.run_without_realign()
            except Exception:
                logging.error(
                    f"Error running {gene} for sample {sample_id}...See error message below"
                )
                traceback.print_exc()
                phaser_calls.setdefault(gene, None)
        return phaser_calls

    def process_sample(
        self,
        bamlist,
        outdir,
        configs,
        tmpdir,
        prog_cmd,
        num_threads=1,
        dcov={},
        novcf=False,
        genome_build="38",
        gene1only=False,
    ):
        """Main workflow"""
        for bam in bamlist:
            try:
                sample_id = bam.split("/")[-1].split(".")[0]
                logging.info(
                    f"Processing sample {sample_id} at {datetime.datetime.now()}..."
                )
                sample_out = {}
                gdepth = None
                sample_sex = None
                query_genes = list(configs.keys())

                logging.info(
                    f"Getting genome depth for sample {sample_id} at {datetime.datetime.now()}..."
                )
                if sample_id in dcov:
                    gdepth = dcov[sample_id]
                if (
                    gdepth is None
                    and set(query_genes) - set(self.no_genome_depth_genes) != set()
                ):
                    depth = GenomeDepth(
                        bam,
                        os.path.join(
                            os.path.dirname(__file__),
                            "data",
                            genome_build,
                            "genome_region.bed",
                        ),
                    )
                    gdepth, gmad = depth.call()
                    if gdepth < 10 or gmad > 0.25:
                        logging.warning(
                            f"For sample {sample_id}, due to low or highly variable genome coverage, genome coverage is not used for depth correction."
                        )
                        gdepth = None
                if set(query_genes).intersection(set(self.check_sex_genes)) != set():
                    logging.info(f"Checking sample sex at {datetime.datetime.now()}...")
                    depth = GenomeDepth(
                        bam,
                        os.path.join(
                            os.path.dirname(__file__),
                            "data",
                            genome_build,
                            "sex_region.bed",
                        ),
                    )
                    sample_sex = depth.check_sex()

                if num_threads == 1:
                    sample_out = self.process_gene(
                        query_genes,
                        configs,
                        outdir,
                        tmpdir,
                        gdepth,
                        bam,
                        sample_sex,
                        novcf,
                        prog_cmd,
                        gene1only,
                    )
                else:
                    process_gene_partial = partial(
                        self.process_gene,
                        configs=configs,
                        outdir=outdir,
                        tmpdir=tmpdir,
                        gdepth=gdepth,
                        bam=bam,
                        sample_sex=sample_sex,
                        novcf=novcf,
                        prog_cmd=prog_cmd,
                        gene1only=gene1only,
                    )
                    gene_groups = [
                        query_genes[i::num_threads] for i in range(num_threads)
                    ]
                    pool = mp.Pool(num_threads)
                    phaser_calls = pool.map(process_gene_partial, gene_groups)
                    pool.close()
                    pool.join()
                    for phaser_call_set in phaser_calls:
                        sample_out.update(phaser_call_set)
                sample_out = dict(sorted(sample_out.items()))

                logging.info(
                    f"Merging all bams for sample {sample_id} at {datetime.datetime.now()}..."
                )
                self.merge_bams(query_genes, outdir, tmpdir, sample_id)

                logging.info(
                    f"Writing to json for sample {sample_id} at {datetime.datetime.now()}..."
                )
                out_json = os.path.join(outdir, sample_id + ".json")
                with open(out_json, "w") as json_output:
                    json.dump(sample_out, json_output, indent=4)
            except Exception:
                logging.error(
                    f"Error running sample {sample_id}...See error message below"
                )
                traceback.print_exc()

    def merge_bams(self, query_genes, outdir, tmpdir, sample_id):
        """Merge realigned tagged bams for each gene into one bam"""
        bams = []
        for gene in query_genes:
            gene_bam = os.path.join(tmpdir, f"{sample_id}_{gene}_realigned_tagged.bam")
            if os.path.exists(gene_bam) is False:
                gene_bam = os.path.join(tmpdir, f"{sample_id}_{gene}_realigned.bam")
            if os.path.exists(gene_bam):
                bams.append(gene_bam)
            if gene in self.two_reference_regions_genes:
                gene2_bam = os.path.join(
                    tmpdir, f"{sample_id}_{gene}_gene2_realigned_tagged.bam"
                )
                if os.path.exists(gene2_bam) is False:
                    gene2_bam = os.path.join(
                        tmpdir, f"{sample_id}_{gene}_gene2_realigned.bam"
                    )
                if os.path.exists(gene2_bam):
                    bams.append(gene2_bam)
        bam_list_file = os.path.join(tmpdir, f"{sample_id}_bam_list.txt")
        with open(bam_list_file, "w") as fout:
            for bam in bams:
                fout.write(bam + "\n")
        merged_bam = os.path.join(outdir, f"{sample_id}_realigned_tagged.bam")
        tmp_bam = os.path.join(tmpdir, f"{sample_id}_merged.bam")
        pysam.merge("-f", "-o", tmp_bam, "-b", bam_list_file)
        pysam.sort("-o", merged_bam, tmp_bam)
        pysam.index(merged_bam)

    def get_samtools_minimap2_path(self, args):
        """Get and check path to third-party tools"""
        samtools_check = [
            a for a in [args.samtools, shutil.which("samtools")] if a is not None
        ]
        samtools_check2 = [a for a in samtools_check if os.path.exists(a)]
        if samtools_check2 == []:
            raise Exception("samtools is not found")
        else:
            self.samtools = samtools_check2[0]

        minimap2_check = [
            a for a in [args.minimap2, shutil.which("minimap2")] if a is not None
        ]
        minimap2_check2 = [a for a in minimap2_check if os.path.exists(a)]
        if minimap2_check2 == []:
            raise Exception("minimap2 is not found")
        else:
            self.minimap2 = minimap2_check2[0]

    def update_config(self, gene_list, ref_dir, genome, genome_build):
        """Get config info for each gene"""
        genome_ref = pysam.FastaFile(genome)
        data_path = os.path.join(os.path.dirname(__file__), "data")
        configs = {}
        for gene in gene_list:
            assert gene in self.region_config_parsed
            configs.setdefault(gene, self.region_config_parsed[gene])
            configs[gene].setdefault("gene", gene)
            # check third-party tools
            configs[gene].setdefault("tools", {})
            configs[gene]["tools"].setdefault("samtools", self.samtools)
            configs[gene]["tools"].setdefault("minimap2", self.minimap2)
            # update paths
            configs[gene].setdefault("data", {})
            data_paths = configs[gene]["data"]
            genome_build_dir = genome_build
            if genome_build == "37":
                genome_build_dir = "19"
            for data_entry in data_paths:
                old_data_file = data_paths[data_entry]
                new_data_file = os.path.join(
                    data_path, genome_build_dir, gene, old_data_file
                )
                data_paths[data_entry] = new_data_file
            # add reference fasta
            ref_file = os.path.join(ref_dir, f"{gene}_ref.fa")
            data_paths.setdefault("reference", ref_file)
            realign_region = configs[gene].get("realign_region")
            if genome_build == "37":
                realign_region = realign_region.replace("chr", "")
                configs[gene]["realign_region"] = realign_region
                extract_regions = configs[gene].get("extract_regions")
                configs[gene]["extract_regions"] = extract_regions.replace("chr", "")
            self.make_ref_fasta(ref_file, realign_region, genome)
            # if gene2 is specified
            gene2_region = configs[gene].get("gene2_region")
            if gene2_region is not None:
                if genome_build == "37":
                    gene2_region = gene2_region.replace("chr", "")
                    configs[gene]["gene2_region"] = gene2_region
                nchr_gene2 = gene2_region.split(":")[0]
                configs[gene].setdefault("nchr_gene2", nchr_gene2)
                nchr_length_gene2 = genome_ref.get_reference_length(nchr_gene2)
                configs[gene].setdefault("nchr_length_gene2", nchr_length_gene2)
                gene2_ref_file = os.path.join(ref_dir, f"{gene}_gene2_ref.fa")
                self.make_ref_fasta(gene2_ref_file, gene2_region, genome)
                data_paths.setdefault("reference_gene2", gene2_ref_file)
                position_match_file = os.path.join(
                    ref_dir, f"{gene}_position_match.txt"
                )
                self.make_position_match_file(
                    position_match_file, gene, realign_region, ref_dir
                )
                data_paths.setdefault("gene_position_match", position_match_file)

            # check files exist
            for data_file in list(data_paths.values()):
                if os.path.exists(data_file) is False:
                    raise Exception(f"File {data_file} not found.")
            nchr = realign_region.split(":")[0]
            nchr_length = genome_ref.get_reference_length(nchr)
            nchr_old = realign_region.replace(":", "_").replace("-", "_")
            configs[gene].setdefault("nchr", nchr)
            configs[gene].setdefault("nchr_length", nchr_length)
            configs[gene].setdefault("nchr_old", nchr_old)
        genome_ref.close()
        return configs

    def make_ref_fasta(self, ref_file, realign_region, genome):
        """Create a reference file for each region of interest for realignment"""
        make_ref_cmd = (
            f"{self.samtools} faidx {genome} {realign_region} | sed -e "
            + f'"s/-/_/"'
            + " | sed -e "
            + f'"s/:/_/"'
            + f" > {ref_file}"
        )
        result = subprocess.run(
            make_ref_cmd, capture_output=True, text=True, shell=True
        )
        result.check_returncode()
        pysam.faidx(ref_file)

    def make_position_match_file(
        self, position_match_file, gene, realign_region, tmpdir
    ):
        """
        For variant calling against a second gene, align both sequences
        and extract matching positions
        """
        tmp_bam = os.path.join(tmpdir, f"{gene}_aln.bam")
        ref_file = os.path.join(tmpdir, f"{gene}_ref.fa")
        gene2_ref_file = os.path.join(tmpdir, f"{gene}_gene2_ref.fa")
        minimap_cmd = f"{self.minimap2} -a {ref_file} {gene2_ref_file} | {self.samtools} view -bS | {self.samtools} sort > {tmp_bam}"
        result = subprocess.run(minimap_cmd, capture_output=True, text=True, shell=True)
        result.check_returncode()
        pysam.index(tmp_bam)
        tmp_bamh = pysam.AlignmentFile(tmp_bam, "rb")
        ref_name = realign_region.replace(":", "_").replace("-", "_")
        ref_name_split = ref_name.split("_")
        ref_name_offset = int(ref_name_split[1])
        with open(position_match_file, "w") as fout:
            for pileupcolumn in tmp_bamh.pileup(ref_name):
                ref_pos = pileupcolumn.pos
                for read in pileupcolumn.pileups:
                    if (
                        not read.is_del
                        and not read.is_refskip
                        and read.alignment.is_supplementary is False
                        and read.alignment.is_secondary is False
                    ):
                        read_name = read.alignment.query_name
                        read_name_split = read_name.split("_")
                        read_name_start = int(read_name_split[1])
                        read_name_end = int(read_name_split[2])
                        read_pos = read.query_position
                        if read.alignment.is_reverse is False:
                            fout.write(
                                str(ref_pos + ref_name_offset)
                                + "\t"
                                + str(read_pos + read_name_start)
                                + "\n"
                            )
                        else:
                            fout.write(
                                str(ref_pos + ref_name_offset)
                                + "\t"
                                + str(read_name_end - read_pos)
                                + "\n"
                            )
                        break
        tmp_bamh.close()

    def get_gene_list(self, gene_input):
        """Get a list of genes to analyze"""
        if gene_input is None:
            if self.genes_to_call is not None and self.genes_to_call != []:
                return self.genes_to_call
        else:
            gene_list = set()
            for gene in gene_input.split(","):
                if gene in self.accepted_genes:
                    gene_list.add(gene)
                else:
                    logging.warning(
                        f"{gene} is not a valid gene name and will be skipped..."
                    )
            return list(gene_list)
        return self.accepted_genes

    def load_parameters(self):
        parser = argparse.ArgumentParser(
            description="Paraphase: HiFi-based caller for highly homologous genes",
            formatter_class=RawTextHelpFormatter,
        )
        inputp = parser.add_argument_group("Input Options")
        outputp = parser.add_argument_group("Output Options")
        input_group = inputp.add_mutually_exclusive_group(required=True)
        input_group.add_argument(
            "-b",
            "--bam",
            help="Input BAM file, mutually exclusive with -l",
        )
        input_group.add_argument(
            "-l",
            "--list",
            help="File listing absolute paths to multiple input BAM files, one per line",
        )
        inputp.add_argument(
            "-r",
            "--reference",
            help=f"Path to reference genome fasta file",
            required=True,
        )
        outputp.add_argument(
            "-o",
            "--out",
            help="Output directory",
            required=True,
        )
        parser.add_argument(
            "-g",
            "--gene",
            help="Optionally specify which gene(s) to run (separated by comma).\n"
            + "Will run all genes if not specified.\n"
            + "The full set of accepted genes are defined in the config file.\n",
            required=False,
        )
        parser.add_argument(
            "-c",
            "--config",
            help="Optional path to a user-defined config file listing the full set of regions to analyze.\n"
            + "By default Paraphase uses the config file in paraphase/data/38/config.yaml",
            required=False,
        )
        parser.add_argument(
            "-d",
            "--depth",
            help="Optional path to a file listing average depth for each sample",
            required=False,
        )
        parser.add_argument(
            "-t",
            "--threads",
            help="Optional number of threads",
            required=False,
            type=int,
            default=1,
        )
        parser.add_argument(
            "--genome",
            help="Optionally specify which genome reference build the input BAM files are aligned against.\n"
            + "Accepted values are 19, 37 and 38. Default is 38.\n"
            + "Note that fewer genes are currently supported in 19/37. See paraphase/data/19/config.yaml",
            required=False,
            default="38",
        )
        parser.add_argument(
            "--novcf",
            help="Optional. If specified, paraphase will not write vcfs",
            required=False,
            action="store_true",
        )
        parser.add_argument(
            "--gene1only",
            help="Optional. If specified, variant calls will be made against the main gene only.\n"
            + "By default, for SMN1, PMS2, STRC, NCF1 and IKBKG, haplotypes are assigned to gene or\n"
            + "paralog/pseudogene, and variants are called against gene or paralog/pseudogene, respectively.\n",
            required=False,
            action="store_true",
        )
        parser.add_argument(
            "--samtools",
            help="Optional path to samtools",
            required=False,
        )
        parser.add_argument(
            "--minimap2",
            help="Optional path to minimap2",
            required=False,
        )
        parser.add_argument(
            "-v", "--version", action="version", version=f"{paraphase.__version__}"
        )
        return parser

    def run(self):
        parser = self.load_parameters()
        args = parser.parse_args()
        num_threads = args.threads
        outdir = args.out
        os.makedirs(outdir, exist_ok=True)
        tmpdir = os.path.join(
            outdir,
            f"tmp_{datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S-%f')}_{str(random.random())}",
        )
        os.makedirs(tmpdir, exist_ok=True)

        try:
            prog_cmd = " ".join(sys.argv[1:])
            # for data files, genome build 37 uses the same ones as genome build 19
            genome_build_dir = args.genome
            if genome_build_dir == "37":
                genome_build_dir = "19"
            self.parse_configs(region_config=args.config, genome_build=genome_build_dir)
            self.get_samtools_minimap2_path(args)
            gene_list = self.get_gene_list(args.gene)
            if gene_list == []:
                all_genes_joined = ",".join(self.accepted_genes)
                raise Exception(
                    f"Please provide valid gene name(s). Accepted genes are {all_genes_joined}."
                )
            configs = self.update_config(gene_list, tmpdir, args.reference, args.genome)

            # parse depth file
            dcov = {}
            if args.depth is not None:
                with open(args.depth) as f:
                    for line in f:
                        split_line = line.split()
                        dcov.setdefault(split_line[0], float(split_line[-1]))

            # process sample(s)
            bamlist = []
            # one bam, multiprocess by gene
            if args.bam is not None:
                if os.path.exists(args.bam) and os.path.exists(args.bam + ".bai"):
                    bamlist = [args.bam]
                    self.process_sample(
                        bamlist,
                        outdir,
                        configs,
                        tmpdir,
                        prog_cmd,
                        num_threads,
                        dcov,
                        args.novcf,
                        genome_build_dir,
                        args.gene1only,
                    )
                else:
                    logging.warning(f"{args.bam} bam or bai file doesn't exist")
            # multiple bams, multiprocess by sample
            elif args.list is not None:
                with open(args.list) as f:
                    for line in f:
                        bam = line[:-1]
                        if os.path.exists(bam) and os.path.exists(bam + ".bai"):
                            bamlist.append(bam)
                        else:
                            logging.warning(f"{bam} bam or bai file doesn't exist")

                process_sample_partial = partial(
                    self.process_sample,
                    outdir=outdir,
                    configs=configs,
                    tmpdir=tmpdir,
                    prog_cmd=prog_cmd,
                    dcov=dcov,
                    novcf=args.novcf,
                    genome_build=genome_build_dir,
                    gene1only=args.gene1only,
                )
                bam_groups = [bamlist[i::num_threads] for i in range(num_threads)]
                pool = mp.Pool(num_threads)
                pool.map(process_sample_partial, bam_groups)
                pool.close()
                pool.join()
            else:
                raise Exception("No input file is given")
        except Exception:
            logging.error("Error running the program...See error message below")
            traceback.print_exc()
        finally:
            # remove temp dir
            if os.path.exists(tmpdir):
                shutil.rmtree(tmpdir)
            logging.info(
                f"Completed Paraphase analysis at {datetime.datetime.now()}..."
            )



import pytest
import yaml
import os
from paraphase.genes.f8_phaser import F8Phaser
from .test_phaser import update_config


class TestF8Phaser(object):
    cur_dir = os.path.dirname(__file__)
    sample_dir = os.path.join(cur_dir, "test_data")

    def test_inversion(self):
        sample_id = "inv"
        genome_bam = os.path.join(self.sample_dir, f"f8_{sample_id}_genome.bam")
        phaser = F8Phaser(
            sample_id, self.sample_dir, genome_bam=genome_bam, sample_sex="male"
        )
        config = update_config("f8")
        phaser.set_parameter(config)
        f8_call = phaser.call()
        assert f8_call.sv_called == {"int22h3_hap1": "inversion"}

    def test_deletion(self):
        sample_id = "del"
        genome_bam = os.path.join(self.sample_dir, f"f8_{sample_id}_genome.bam")
        phaser = F8Phaser(
            sample_id, self.sample_dir, genome_bam=genome_bam, sample_sex="male"
        )
        config = update_config("f8")
        phaser.set_parameter(config)
        f8_call = phaser.call()
        assert f8_call.sv_called == {"int22h2_hap1": "deletion"}



# paraphase
# Author: Xiao Chen <xchen@pacificbiosciences.com>


import os
import pysam
import subprocess
import random
import re
import math
from scipy.stats import poisson
from collections import Counter
from .haplotype_assembler import VariantGraph
import paraphase


class BamRealigner:
    """
    Extract and realign reads to region of interest
    """

    min_mapq = 50
    min_aln = 800
    max_mismatch = 0.05
    deletion = r"\d+D"
    insertion = r"\d+I"

    def __init__(self, bam, outdir, config, prog_cmd):
        self.bam = bam
        self.outdir = outdir
        self.prog_cmd = prog_cmd
        self.gene = config["gene"]
        self.ref = config["data"]["reference"]
        self.nchr_old = config["nchr_old"]
        self.nchr = config["nchr"]
        self.offset = int(self.nchr_old.split("_")[1]) - 1
        self.nchr_length = config["nchr_length"]
        self.extract_regions = config["extract_regions"]
        self.samtools = config["tools"]["samtools"]
        self.minimap2 = config["tools"]["minimap2"]
        if "check_nm" in config:
            self.max_mismatch = config["check_nm"]
        self.use_r2k = ""
        if "use_r2k" in config:
            self.use_r2k = "-r2k"
        self._bamh = pysam.AlignmentFile(bam, "rb")
        self.sample_id = bam.split("/")[-1].split(".")[0]
        self.realign_bam = os.path.join(
            self.outdir, self.sample_id + f"_{self.gene}_realigned_old.bam"
        )
        self.realign_out_bam = os.path.join(
            self.outdir, self.sample_id + f"_{self.gene}_realigned.bam"
        )
        self.gene2_region = config.get("gene2_region")
        if self.gene2_region is not None:
            self.nchr_gene2 = config["nchr_gene2"]
            self.offset_gene2 = int(self.gene2_region.split(":")[1].split("-")[0]) - 1
            self.nchr_length_gene2 = config["nchr_length_gene2"]
            self.gene2_ref = config["data"].get("reference_gene2")
            self.realign_bam_gene2 = os.path.join(
                self.outdir, self.sample_id + f"_{self.gene}_gene2_realigned_old.bam"
            )
            self.realign_out_bam_gene2 = os.path.join(
                self.outdir, self.sample_id + f"_{self.gene}_gene2_realigned.bam"
            )

    def get_nm(self, read, min_size=300):
        """Get number of mismatches excluding big deletions or insertions"""
        cigar = read.cigarstring
        deletions_on_read = [int(a[:-1]) for a in re.findall(self.deletion, cigar)]
        large_deletions = [a for a in deletions_on_read if a > min_size]
        insertions_on_read = [int(a[:-1]) for a in re.findall(self.insertion, cigar)]
        large_insertions = [a for a in insertions_on_read if a > min_size]
        return read.get_tag("NM") - sum(large_deletions + large_insertions)

    def write_realign_bam(self, gene2=False):
        """
        Realign reads to region of interest and output a tagged bam for visualization
        """
        if gene2 is False:
            realign_ref = self.ref
            realign_out_tmp = self.realign_bam
            realign_out = self.realign_out_bam
            ref_name_tmp = self.nchr_old
            offset = self.offset
            nchr = self.nchr
            nchr_length = self.nchr_length
        else:
            realign_ref = self.gene2_ref
            realign_out_tmp = self.realign_bam_gene2
            realign_out = self.realign_out_bam_gene2
            ref_name_tmp = self.gene2_region.replace(":", "_").replace("-", "_")
            offset = self.offset_gene2
            nchr = self.nchr_gene2
            nchr_length = self.nchr_length_gene2

        realign_cmd = (
            f"{self.samtools} view -F 0x100 -F 0x200 -F 0x800 {self.bam} {self.extract_regions} | sort | uniq | "
            + f'awk \'BEGIN {{FS="\\t"}} {{print "@" $1 "\\n" $10 "\\n+\\n" $11}}\''
            + f" | {self.minimap2} {self.use_r2k} -a -x map-pb {realign_ref} - | {self.samtools} view -bh | {self.samtools} sort > {realign_out_tmp}"
        )
        result = subprocess.run(realign_cmd, capture_output=True, text=True, shell=True)
        result.check_returncode()
        if os.path.exists(realign_out_tmp) is False:
            raise Exception("Realigned bam does not exist.")

        pysam.index(realign_out_tmp)
        realign_bamh = pysam.AlignmentFile(realign_out_tmp, "rb")
        realign_bamh_header = realign_bamh.header
        realign_bamh_header = realign_bamh_header.to_dict()
        pg_lines = []
        pg_lines_original = realign_bamh_header.get("PG")
        if pg_lines_original is not None:
            pg_lines = [
                a for a in pg_lines_original if "ID" in a and a["ID"] == "minimap2"
            ]
        pg_lines.append(
            {
                "PN": "paraphase",
                "ID": "paraphase",
                "VN": paraphase.__version__,
                "CL": f"paraphase {self.prog_cmd}",
            }
        )
        new_header = {
            "SQ": [{"SN": nchr, "LN": nchr_length}],
            "PG": pg_lines,
        }
        realign_out_bamh = pysam.AlignmentFile(
            realign_out,
            "wb",
            header=new_header,
        )
        for read in realign_bamh.fetch(ref_name_tmp):
            num_mismatch = self.get_nm(read)
            if (
                read.mapping_quality >= self.min_mapq
                and read.query_alignment_length >= self.min_aln
                and num_mismatch < read.reference_length * self.max_mismatch
            ):
                read.reference_start += offset
                ltags = read.tags
                new_ltags = []
                for tag in ltags:
                    if tag[0] != "SA":
                        new_ltags.append(tag)
                    else:
                        # update the SA tag
                        new_sa = []
                        for sa_item in tag[1].split(";"):
                            if sa_item != "":
                                at = sa_item.split(",")
                                at[0] = nchr
                                tmp_pos = int(at[1])
                                at[1] = str(tmp_pos + offset)
                                mapq = int(at[4])
                                if mapq >= self.min_mapq:
                                    new_at = ",".join(at)
                                    new_sa.append(new_at)
                        if new_sa != []:
                            new_sa.append("")
                            new_ltags.append(("SA", ";".join(new_sa)))
                read.tags = new_ltags
                realign_out_bamh.write(read)
        realign_bamh.close()
        realign_out_bamh.close()
        pysam.index(realign_out)
        os.remove(realign_out_tmp)
        os.remove(realign_out_tmp + ".bai")


class BamTagger:
    """
    Add HP tags to bam
    """

    read_color = "166,206,227"
    read_color_allele1 = "178,223,138"
    read_color_allele2 = "177,156,217"

    def __init__(self, sample_id, outdir, config, call_sum):
        self.sample_id = sample_id
        self.outdir = outdir
        self.call_sum = call_sum
        self.gene = config["gene"]
        self.bam = os.path.join(
            self.outdir, self.sample_id + f"_{self.gene}_realigned.bam"
        )
        self.nchr = config["nchr"]
        self.tmp_bam = os.path.join(
            self.outdir, self.sample_id + f"_{self.gene}_tmp.bam"
        )
        self.tagged_realigned_bam = os.path.join(
            self.outdir, self.sample_id + f"_{self.gene}_realigned_tagged.bam"
        )
        self.use_supplementary = False
        if "use_supplementary" in config or "is_tandem" in config:
            self.use_supplementary = True
        self.gene2_region = config.get("gene2_region")
        if self.gene2_region is not None:
            self.nchr_gene2 = config["nchr_gene2"]
            self.bam_gene2 = os.path.join(
                self.outdir, self.sample_id + f"_{self.gene}_gene2_realigned.bam"
            )
            self.tmp_bam_gene2 = os.path.join(
                self.outdir, self.sample_id + f"_{self.gene}_gene2_tmp.bam"
            )
            self.tagged_realigned_bam_gene2 = os.path.join(
                self.outdir, self.sample_id + f"_{self.gene}_gene2_realigned_tagged.bam"
            )
        random.seed(0)

    def add_tag_to_read(
        self,
        read,
        hp_keys,
        reads_to_tag,
        nonunique,
        read_details,
        alleles=[],
        random_assign=False,
        gene2=False,
    ):
        """
        Add HP tag to each read.
        Option to randomly assign a read to a possible haplotype when
        nonuniquelly matching. The randomly assigned reads are in gray
        while unique reads are in blue.
        """
        hp_found = False
        read_name = read.qname
        if (
            read.is_supplementary
            and self.use_supplementary
            and gene2 is False
            # coordinates are all changed after realigning to gene2
        ):
            read_name = (
                read_name + f"_sup_{read.reference_start}_{read.reference_length}"
            )
        for hap, hap_name in hp_keys.items():
            if read_name in reads_to_tag[hap]:
                read.set_tag("HP", hap_name, "Z")
                read.set_tag("YC", self.read_color, "Z")
                if alleles != []:
                    if hap_name in alleles[0]:
                        read.set_tag("YC", self.read_color_allele1, "Z")
                    elif len(alleles) > 1 and hap_name in alleles[1]:
                        read.set_tag("YC", self.read_color_allele2, "Z")
                hp_found = True
        if hp_found is False:
            # find closest match
            if read_name in read_details:
                read_seq = read_details[read_name]
                keys = []
                mismatches = []
                for ass_hap in hp_keys.keys():
                    match, mismatch, extend = VariantGraph.compare_two_haps(
                        read_seq, ass_hap
                    )
                    keys.append(ass_hap)
                    mismatches.append(mismatch)
                mismatches_sorted = sorted(mismatches)
                if (
                    len(mismatches_sorted) > 1
                    and 0 < mismatches_sorted[0] <= 2
                    and mismatches_sorted[1] >= mismatches_sorted[0] + 2
                ):
                    best_match = keys[mismatches.index(mismatches_sorted[0])]
                    read.set_tag("HP", hp_keys[best_match], "Z")
                    hp_found = True
        if hp_found is False:
            if random_assign is False:
                read.set_tag("HP", "Unassigned", "Z")
            else:
                if read_name in nonunique:
                    possible_haps = nonunique[read_name]
                    random_hap = possible_haps[
                        random.randint(0, len(possible_haps) - 1)
                    ]
                    if random_hap in hp_keys:
                        read.set_tag("HP", hp_keys[random_hap], "Z")
                else:
                    read.set_tag("HP", "Unassigned", "Z")
        return read

    def write_bam(self, random_assign=False, gene2=False):
        """
        Write HP tags in output bams.
        """
        if gene2 is False:
            inbam = self.bam
            tmpbam = self.tmp_bam
            outbam = self.tagged_realigned_bam
            nchr = self.nchr
        else:
            inbam = self.bam_gene2
            tmpbam = self.tmp_bam_gene2
            outbam = self.tagged_realigned_bam_gene2
            nchr = self.nchr_gene2

        call_sum = self.call_sum
        hp_keys = call_sum.get("final_haplotypes")
        if hp_keys is None:
            return
        self._bamh = pysam.AlignmentFile(inbam, "rb")
        out_bamh = pysam.AlignmentFile(tmpbam, "wb", template=self._bamh)
        unique_reads = call_sum.get("unique_supporting_reads")
        nonunique_reads = call_sum.get("nonunique_supporting_reads")
        if nonunique_reads is None:
            nonunique_reads = {}
        read_details = call_sum.get("read_details")
        alleles = call_sum.get("alleles_final")
        if alleles is None:
            alleles = []

        for read in self._bamh.fetch(nchr):
            if read.is_secondary is False:
                read = self.add_tag_to_read(
                    read,
                    hp_keys,
                    unique_reads,
                    nonunique_reads,
                    read_details,
                    alleles,
                    random_assign=random_assign,
                    gene2=gene2,
                )
                out_bamh.write(read)
        out_bamh.close()
        self._bamh.close()
        pysam.sort("-o", outbam, tmpbam)
        pysam.index(outbam)
        os.remove(tmpbam)
        os.remove(inbam)
        os.remove(inbam + ".bai")


class VcfGenerater:
    """
    Call variants and generate individual/merged vcfs
    """

    search_range = 200

    def __init__(self, sample_id, outdir, call_sum):
        self.sample_id = sample_id
        self.outdir = outdir
        self.call_sum = call_sum
        self.match = {}

    def set_parameter(self, config, tmpdir=None, prog_cmd=None):
        self.gene = config["gene"]
        self.nchr = config["nchr"]
        self.nchr_old = config["nchr_old"]
        self.offset = int(self.nchr_old.split("_")[1]) - 1
        self.nchr_length = config["nchr_length"]
        self.ref = config["data"]["reference"]
        self.left_boundary = config.get("left_boundary")
        self.right_boundary = config.get("right_boundary")
        self.deletion1_in_gene1 = config.get("deletion1_in_gene1")
        self.deletion1_in_gene2 = config.get("deletion1_in_gene2")
        self.extract_regions = config.get("extract_regions")
        if self.left_boundary is None:
            self.left_boundary = int(self.nchr_old.split("_")[1])
        if self.right_boundary is None:
            self.right_boundary = int(self.nchr_old.split("_")[2])
        self.samtools = config["tools"]["samtools"]
        self.minimap2 = config["tools"]["minimap2"]
        self.use_supplementary = False
        if "use_supplementary" in config or "is_tandem" in config:
            self.use_supplementary = True
        self.keep_truncated = False
        if "keep_truncated" in config:
            self.keep_truncated = True

        self.prog_cmd = prog_cmd
        self.tmpdir = tmpdir
        if self.tmpdir is None:
            self.tmpdir = self.outdir
        self.bam = os.path.join(
            tmpdir, self.sample_id + f"_{self.gene}_realigned_tagged.bam"
        )
        self.vcf_dir = os.path.join(self.outdir, f"{self.sample_id}_vcfs")
        os.makedirs(self.vcf_dir, exist_ok=True)

    def get_range_in_other_gene(self, pos):
        """
        Find the correponding coordinates in the other gene
        """
        if pos in self.match:
            return self.match[pos]
        for i in range(self.search_range):
            new_pos = pos + i
            if new_pos in self.match:
                return self.match[new_pos]
        return None

    def write_header(self, fout):
        """Write VCF header"""
        fout.write("##fileformat=VCFv4.2\n")
        fout.write('##FILTER=<ID=PASS,Description="All filters passed">\n')
        fout.write('##FILTER=<ID=LowDP,Description="Low depth at this site.">\n')
        fout.write(
            '##FILTER=<ID=LowQual,Description="Low confidence in this variant.">\n'
        )
        fout.write(
            '##INFO=<ID=HapIDs,Number=R,Type=String,Description="Haplotype IDs">\n'
        )
        if self.gene in ["ikbkg", "f8"]:
            fout.write(
                '##INFO=<ID=SVLEN,Number=1,Type=Integer,Description="Length of the SV">\n'
            )
            fout.write(
                '##INFO=<ID=SVTYPE,Number=1,Type=String,Description="Type of the SV.">\n'
            )
            fout.write(
                '##INFO=<ID=END,Number=1,Type=Integer,Description="End position of the structural variant">\n'
            )
            fout.write('##ALT=<ID=DEL,Description="Deletion">\n')
            fout.write('##ALT=<ID=INV,Description="Inversion">\n')
        fout.write('##FORMAT=<ID=GT,Number=R,Type=String,Description="Genotype">\n')
        fout.write('##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read depth">\n')
        fout.write(
            '##FORMAT=<ID=AD,Number=R,Type=Integer,Description="Read depth for each allele">\n'
        )
        fout.write(f"##contig=<ID={self.nchr},length={self.nchr_length}>\n")
        fout.write(f"##paraphase_version={paraphase.__version__}\n")
        fout.write(f"##paraphase_command=paraphase {self.prog_cmd}\n")
        header = [
            "#CHROM",
            "POS",
            "ID",
            "REF",
            "ALT",
            "QUAL",
            "FILTER",
            "INFO",
            "FORMAT",
            "default",
        ]
        fout.write("\t".join(header) + "\n")

    def merge_vcf(self, vars_list):
        """
        Merge vcfs from multiple haplotypes.
        """
        merged_vcf = os.path.join(
            self.vcf_dir, self.sample_id + f"_{self.gene}_variants.vcf"
        )
        with open(merged_vcf, "w") as fout:
            self.write_header(fout)
            for variants_info, haps_ids in vars_list:
                variants_info = dict(sorted(variants_info.items()))
                for pos in variants_info:
                    call_info = variants_info[pos]
                    # unique variants at this site
                    variant_observed = set([a[0] for a in call_info if a is not None])
                    var_num = len(variant_observed)
                    for variant in variant_observed:
                        _, ref, alt = variant.split("_")
                        merge_gt = []
                        merge_ad = []
                        merge_dp = []
                        for each_call in call_info:
                            if each_call is None:
                                merge_gt.append(".")
                                merge_ad.append(".")
                                merge_dp.append(".")
                            else:
                                var_name, dp, ad, var_filter, gt = each_call
                                if var_filter != []:
                                    gt = "."
                                merge_dp.append(str(dp))
                                if gt == "0":
                                    merge_gt.append(gt)
                                    merge_ad.append(str(dp - ad))
                                elif var_name == variant:
                                    merge_gt.append(gt)
                                    merge_ad.append(str(ad))
                                else:
                                    merge_gt.append(".")
                                    merge_ad.append(".")
                        final_qual = "."
                        if ref == alt:
                            alt = "."
                        if (var_num == 1 and ("1" in merge_gt or "." in merge_gt)) or (
                            var_num > 1 and alt != "."
                        ):
                            if alt.isdigit() is False:
                                merged_entry = [
                                    self.nchr,
                                    str(pos),
                                    ".",
                                    ref,
                                    alt,
                                    final_qual,
                                    "PASS",
                                    "HapIDs=" + ",".join(haps_ids),
                                    "GT:DP:AD",
                                    "/".join(merge_gt)
                                    + ":"
                                    + ",".join(merge_dp)
                                    + ":"
                                    + ",".join(merge_ad),
                                ]
                            else:
                                nstart, var_type, nend = variant.split("_")
                                nstart = int(nstart)
                                nend = int(nend)
                                var_size = nend - nstart
                                merged_entry = [
                                    self.nchr,
                                    str(pos),
                                    ".",
                                    "N",
                                    f"<{var_type}>",
                                    final_qual,
                                    "PASS",
                                    f"SVTYPE={var_type};END={nend};SVLEN={var_size};"
                                    + "HapIDs="
                                    + ",".join(haps_ids),
                                    "GT:DP:AD",
                                    "/".join(merge_gt)
                                    + ":"
                                    + ",".join(merge_dp)
                                    + ":"
                                    + ",".join(merge_ad),
                                ]
                            fout.write("\t".join(merged_entry) + "\n")

    @staticmethod
    def refine_indels(ref_seq, var_seq, pos, refh, ref_name):
        """process indels"""
        if "+" in var_seq:
            ins_base = var_seq.split(re.findall(r"\+\d+", var_seq)[0])[1]
            var_seq = ref_seq + ins_base
        elif "-" in var_seq:
            del_len = int(re.findall(r"\-\d+", var_seq)[0][1:])
            var_seq = ref_seq
            ref_seq = refh.fetch(
                ref_name,
                pos - 1,
                pos + del_len,
            )
        return ref_seq, var_seq

    @staticmethod
    def get_var_call_qual(dp, ad, gt, is_snp):
        "Calculate variant call quality. Not used"
        noise_ref = 0.03
        noise_alt = 0.05
        if is_snp is False:
            noise_alt = 0.1
        min_qual = 0
        max_qual = 100
        if gt == "0":
            # non-allele reads given noise
            h0 = poisson.pmf(dp - ad, dp * noise_ref)
            # allele reads given noise
            h1 = poisson.pmf(ad, dp * noise_alt)
        if gt == "1":
            # non-allele reads given noise
            h0 = poisson.pmf(dp - ad, dp * noise_alt)
            # allele reads given noise
            h1 = poisson.pmf(ad, dp * noise_alt)
        qual = math.floor(10 * math.log(h0 / h1, 10))
        return max(min(qual, max_qual), min_qual)

    @staticmethod
    def get_var(all_bases, ref_seq):
        """Get most supported base as variant"""
        dp = len(all_bases)
        gt = "."
        qual = "."
        ad = len([a for a in all_bases if a != ref_seq])
        var_seq = ref_seq
        if all_bases != []:
            counter = Counter(all_bases)
            most_common_base = counter.most_common(2)
            var_seq = most_common_base[0][0]
            ad = most_common_base[0][1]
            is_snp = False
            if len(var_seq) == 1 and len(ref_seq) == 1 and var_seq != "*":
                is_snp = True
            if var_seq == ref_seq:
                gt = "0"
            else:
                gt = "1"
            # qual = VcfGenerater.get_var_call_qual(dp, ad, gt, is_snp)
        return [var_seq, dp, ad, gt, qual]

    def get_hap_bound(self, hap_name):
        """Get haplotype boundaries"""
        hap_bound = list(self.call_sum["haplotype_details"][hap_name]["boundary"])
        # find the positions next to the existing boundaries
        confident_position = hap_bound[0]
        if confident_position == self.left_boundary:
            hap_bound.append(confident_position)
        else:
            for var in self.call_sum["sites_for_phasing"]:
                confident_position = int(var.split("_")[0])
                if confident_position > hap_bound[0]:
                    break
            hap_bound.append(confident_position)
        confident_position = hap_bound[1]
        if confident_position == self.right_boundary:
            hap_bound.append(confident_position)
        else:
            for var in reversed(self.call_sum["sites_for_phasing"]):
                confident_position = int(var.split("_")[0])
                if confident_position < hap_bound[1]:
                    break
            hap_bound.append(confident_position)
        return hap_bound

    def pileup_to_variant(
        self,
        pileups_raw,
        read_names,
        uniq_reads,
        refh,
        offset,
        hap_bound,
        vcf_out,
        min_depth=4,
        min_qual=25,
        variants_to_add={},
    ):
        """
        Filter pileups and make variant calls.
        """
        variants = []
        # for now, report F8 SVs in the haplotype vcfs. Ideally they should be in a diploid vcf.
        if self.gene == "f8":
            for pos in variants_to_add:
                var_name = variants_to_add[pos]
                nstart, var_type, nend = var_name.split("_")
                nstart = int(nstart)
                nend = int(nend)
                var_size = nend - nstart
                variants.append([nstart, var_name, ".", ".", [], "1"])
                vcf_out_line = [
                    self.nchr,
                    str(nstart),
                    ".",
                    "N",
                    f"<{var_type}>",
                    ".",
                    "PASS",
                    f"SVTYPE={var_type};END={nend};SVLEN={var_size}",
                    "GT:DP:AD",
                    f"1:.:.",
                ]
                vcf_out.write("\t".join(vcf_out_line) + "\n")

        ref_name = refh.references[0]
        del_pos = []
        for pos in pileups_raw:
            if pos in variants_to_add:
                var_name = variants_to_add[pos]
                nstart, var_type, nend = var_name.split("_")
                nstart = int(nstart)
                nend = int(nend)
                var_size = nend - nstart
                variants.append([nstart, var_name, ".", ".", [], "1"])
                vcf_out_line = [
                    self.nchr,
                    str(nstart),
                    ".",
                    "N",
                    f"<{var_type}>",
                    ".",
                    "PASS",
                    f"SVTYPE={var_type};END={nend};SVLEN={var_size}",
                    "GT:DP:AD",
                    f"1:.:.",
                ]
                vcf_out.write("\t".join(vcf_out_line) + "\n")
                del_pos = [nstart, nend]

            all_bases = pileups_raw[pos]
            if offset < 0:
                true_pos = pos
                refh_pos = pos + offset
            else:
                true_pos = pos + offset
                refh_pos = pos
            ref_seq = refh.fetch(ref_name, refh_pos - 1, refh_pos)
            alt_all_reads = self.get_var(all_bases, ref_seq)
            if (
                hap_bound == []
                or (None not in hap_bound and hap_bound[0] < true_pos < hap_bound[1])
            ) and (del_pos == [] or true_pos < del_pos[0] or true_pos > del_pos[1]):
                # use only unique reads for positions at the edge
                if (
                    hap_bound == []
                    or true_pos < hap_bound[2]
                    or true_pos > hap_bound[3]
                ):
                    bases_uniq_reads = []
                    for i, read_base in enumerate(all_bases):
                        if uniq_reads is None or read_names[pos][i] in uniq_reads:
                            bases_uniq_reads.append(read_base)
                    alt_uniq_reads = self.get_var(bases_uniq_reads, ref_seq)
                    # if alt_uniq_reads[1] >= min_depth:
                    var_seq, dp, ad, gt, qual = alt_uniq_reads
                    # else:
                    #    var_seq, dp, ad, gt, qual = alt_all_reads
                    #    gt = "."
                else:
                    var_seq, dp, ad, gt, qual = alt_all_reads

                ref_seq, var_seq = self.refine_indels(
                    ref_seq, var_seq, refh_pos, refh, ref_name
                )
                var = f"{true_pos}_{ref_seq}_{var_seq}"
                qual = "."
                var_filter = []
                if dp < min_depth:
                    var_filter.append("LowDP")
                # if qual != "." and qual < min_qual:
                if ad < dp * 0.7:
                    var_filter.append("LowQual")
                if var_filter == []:
                    call_filter = "PASS"
                else:
                    call_filter = ";".join(var_filter)
                    gt = "."
                variants.append([true_pos, var, dp, ad, var_filter, gt])
                if var_seq == ref_seq:
                    var_seq = "."
                # write all positions where gt is not confidently 0
                if gt == "1" or gt == ".":
                    vcf_out_line = [
                        self.nchr,
                        str(true_pos),
                        ".",
                        ref_seq,
                        var_seq,
                        str(qual),
                        call_filter,
                        ".",
                        "GT:DP:AD",
                        f"{gt}:{dp}:{ad}",
                    ]
                    vcf_out.write("\t".join(vcf_out_line) + "\n")
        return variants

    def run_without_realign(
        self,
        gene2=False,
        final_haps={},
        match_range=False,
    ):
        """
        Make vcf from existing alignment.
        This works for gene/pseudogene scenarios,
        i.e. no need to realign to pseudogene reference.
        """
        call_sum = self.call_sum
        if gene2 is False:
            final_haps = call_sum.get("final_haplotypes")
        if final_haps is None:
            return
        # special SV type variants to report for certain genes
        special_variants = {}
        if self.gene == "ikbkg":
            del_haps = call_sum.get("deletion_haplotypes")
            if del_haps is not None and del_haps != []:
                for del_hap in del_haps:
                    del_name = self.deletion1_in_gene1
                    if "pseudo" in del_hap and gene2 is True:
                        del_name = self.deletion1_in_gene2
                    special_variants.setdefault(del_hap, del_name)
        if self.gene == "f8":
            sv_called = call_sum.get("sv_called")
            if sv_called is not None and sv_called != {}:
                (
                    extract_region1,
                    extract_region2,
                    extract_region3,
                ) = self.extract_regions.split()
                extract_region1_end = extract_region1.split("-")[1]
                extract_region2_start = extract_region2.split(":")[1].split("-")[0]
                extract_region3_start = extract_region3.split(":")[1].split("-")[0]
                for sv_hap, sv in sv_called.items():
                    if sv == "inversion":
                        sv_name = f"{extract_region1_end}_INV_{extract_region3_start}"
                        special_variants.setdefault(sv_hap, sv_name)
                    elif sv == "deletion":
                        sv_name = f"{extract_region1_end}_DEL_{extract_region2_start}"
                        special_variants.setdefault(sv_hap, sv_name)

        uniq_reads = []
        for read_set in self.call_sum["unique_supporting_reads"].values():
            for read_name in read_set:
                read_name_split = read_name.split("_sup")
                # supplementary alignments
                if self.use_supplementary and len(read_name_split) > 1:
                    uniq_reads.append(read_name_split[0])
                else:
                    uniq_reads.append(read_name)
        variants_info = {}
        two_cp_haplotypes = self.call_sum.get("two_copy_haplotypes")
        # exclude truncated copies
        haps_not_truncated = [
            a
            for a in final_haps.values()
            if self.call_sum["haplotype_details"][a]["is_truncated"] is False
            or self.keep_truncated is True
        ]
        nhap = len(haps_not_truncated) + len(
            [a for a in two_cp_haplotypes if a in haps_not_truncated]
        )
        hap_ids = []

        # gene1only, or two-gene mode but gene1 side
        if gene2 is False or match_range is False:
            bamh = pysam.AlignmentFile(self.bam, "rb")
            refh = pysam.FastaFile(self.ref)
            offset = self.offset
            nchr = self.nchr
        else:
            bamh = pysam.AlignmentFile(self.bam_gene2, "rb")
            refh = pysam.FastaFile(self.ref_gene2)
            offset = self.offset_gene2
            nchr = self.nchr_gene2

        if (gene2 is False or match_range is False) and final_haps == {}:
            hap_name = "homozygous_hap1"
            hap_vcf_out = os.path.join(
                self.vcf_dir, self.sample_id + f"_{self.gene}_{hap_name}.vcf"
            )
            vcf_out = open(hap_vcf_out, "w")
            self.write_header(vcf_out)
            pileups_raw = {}
            read_names = {}
            for pileupcolumn in bamh.pileup(
                nchr,
                truncate=True,
                min_base_quality=30,
            ):
                pos = pileupcolumn.pos + 1
                this_pos_bases = [
                    a.upper() for a in pileupcolumn.get_query_sequences(add_indels=True)
                ]
                pileups_raw.setdefault(pos, this_pos_bases)
                read_names.setdefault(pos, pileupcolumn.get_query_names())
            variants_called = self.pileup_to_variant(
                pileups_raw,
                read_names,
                None,
                refh,
                0 - offset,
                [],
                vcf_out,
            )
            vcf_out.close()
            hap_ids.append(hap_name)
            hap_ids.append(hap_name)
            for pos, var_name, dp, ad, var_filter, gt in variants_called:
                variants_info.setdefault(
                    pos,
                    [
                        [var_name, dp, ad, var_filter, gt],
                        [var_name, dp, ad, var_filter, gt],
                    ],
                )

        i = 0
        for hap_name in final_haps.values():
            if (
                self.call_sum["haplotype_details"][hap_name]["is_truncated"] is True
                and self.keep_truncated is False
            ):
                continue
            hap_ids.append(hap_name)

            variants_to_add = {}
            if hap_name in special_variants:
                variant_to_add = special_variants[hap_name]
                variants_to_add.setdefault(
                    int(variant_to_add.split("_")[0]), variant_to_add
                )

            hap_bound = self.get_hap_bound(hap_name)
            # convert to positions in the other gene
            if match_range:
                if self.match == {}:
                    hap_bound = []
                else:
                    n1 = self.get_range_in_other_gene(hap_bound[0])
                    n2 = self.get_range_in_other_gene(hap_bound[1])
                    n3 = self.get_range_in_other_gene(hap_bound[2])
                    n4 = self.get_range_in_other_gene(hap_bound[3])
                    if None in [n1, n2, n3, n4]:
                        hap_bound = []
                    else:
                        hap_bound = [
                            min(n1, n2),
                            max(n1, n2),
                            min(n3, n4),
                            max(n3, n4),
                        ]
            hap_vcf_out = os.path.join(
                self.vcf_dir, self.sample_id + f"_{self.gene}_{hap_name}.vcf"
            )
            vcf_out = open(hap_vcf_out, "w")
            self.write_header(vcf_out)

            # by HP tag
            pileups_raw = {}
            read_names = {}
            for pileupcolumn in bamh.pileup(
                nchr,
                truncate=True,
                min_base_quality=30,
            ):
                pos = pileupcolumn.pos + 1
                this_pos_bases = [
                    a.upper() for a in pileupcolumn.get_query_sequences(add_indels=True)
                ]
                this_position_hps = []
                this_pos_read_names = []
                this_pos_read_names_sup = []

                for read in pileupcolumn.pileups:
                    read_tag = read.alignment.get_tag("HP")
                    this_position_hps.append(read_tag)
                    read_name = read.alignment.query_name
                    new_read_name = read_name
                    if self.use_supplementary and read.alignment.is_supplementary:
                        new_read_name = (
                            read_name + f"_sup_{read.alignment.reference_start}"
                        )
                    this_pos_read_names.append(read_name)
                    this_pos_read_names_sup.append(new_read_name)

                for read_num, read_hap in enumerate(this_position_hps):
                    if read_hap == hap_name:
                        pileups_raw.setdefault(pos, []).append(this_pos_bases[read_num])
                        if self.use_supplementary:
                            read_names.setdefault(pos, []).append(
                                this_pos_read_names_sup[read_num]
                            )
                        else:
                            read_names.setdefault(pos, []).append(
                                this_pos_read_names[read_num]
                            )

            variants_called = self.pileup_to_variant(
                pileups_raw,
                read_names,
                uniq_reads,
                refh,
                0 - offset,
                hap_bound,
                vcf_out,
                variants_to_add=variants_to_add,
            )
            vcf_out.close()

            for pos, var_name, dp, ad, var_filter, gt in variants_called:
                variants_info.setdefault(pos, [None] * nhap)
                variants_info[pos][i] = [var_name, dp, ad, var_filter, gt]
                if hap_name in two_cp_haplotypes:
                    variants_info[pos][i + 1] = [var_name, dp, ad, var_filter, gt]
            if hap_name in two_cp_haplotypes:
                i += 1
                hap_ids.append(hap_name)
            i += 1

        bamh.close()
        refh.close()
        if gene2 is False:
            self.merge_vcf([(variants_info, hap_ids)])
        else:
            return variants_info, hap_ids


class TwoGeneVcfGenerater(VcfGenerater):
    """
    Make vcf for two-gene scenario
    """

    def __init__(self, sample_id, outdir, call_sum):
        VcfGenerater.__init__(self, sample_id, outdir, call_sum)

    def set_parameter(self, config, tmpdir=None, prog_cmd=None):
        super().set_parameter(config, tmpdir, prog_cmd)
        self.gene2_region = config["gene2_region"]
        self.nchr_gene2 = config["nchr_gene2"]
        self.offset_gene2 = int(self.gene2_region.split(":")[1].split("-")[0]) - 1
        self.ref_gene2 = config["data"]["reference_gene2"]
        self.bam_gene2 = os.path.join(
            tmpdir, self.sample_id + f"_{self.gene}_gene2_realigned_tagged.bam"
        )
        self.position_match = config["data"].get("gene_position_match")
        if self.position_match is not None:
            with open(self.position_match) as f:
                for line in f:
                    at = line.split()
                    self.match.setdefault(int(at[0]), int(at[1]))

    def separate_two_genes(self):
        """Get haplotypes for gene1 and gene2"""
        all_haps = self.call_sum.get("final_haplotypes")
        gene1_haps = {}
        gene2_haps = {}
        if self.gene == "smn1":
            return self.call_sum["smn1_haplotypes"], {
                **self.call_sum["smn2_haplotypes"],
                **self.call_sum["smn2_del78_haplotypes"],
            }
        elif self.gene == "pms2":
            for hap, hap_name in all_haps.items():
                if "pms2cl" not in hap_name:
                    gene1_haps.setdefault(hap, hap_name)
                else:
                    gene2_haps.setdefault(hap, hap_name)
        elif self.gene == "strc":
            for hap, hap_name in all_haps.items():
                if "strcp1" not in hap_name:
                    gene1_haps.setdefault(hap, hap_name)
                else:
                    gene2_haps.setdefault(hap, hap_name)
        elif self.gene == "ncf1":
            for hap, hap_name in all_haps.items():
                if "pseudo" not in hap_name:
                    gene1_haps.setdefault(hap, hap_name)
                else:
                    gene2_haps.setdefault(hap, hap_name)
        elif self.gene == "ikbkg":
            for hap, hap_name in all_haps.items():
                if "dup" not in hap_name:
                    if "pseudo" not in hap_name:
                        gene1_haps.setdefault(hap, hap_name)
                    else:
                        gene2_haps.setdefault(hap, hap_name)
        return gene1_haps, gene2_haps

    def run(self):
        """
        Process haplotypes one by one. Realign to different ref sequence
        in this two-gene scenario
        """
        call_sum = self.call_sum
        if call_sum.get("final_haplotypes") is None:
            return
        gene1_haps, gene2_haps = self.separate_two_genes()
        vars_gene1, gene1_hap_ids = self.run_without_realign(
            gene2=True,
            final_haps=gene1_haps,
        )
        vars_gene2, gene2_hap_ids = self.run_without_realign(
            gene2=True,
            final_haps=gene2_haps,
            match_range=True,
        )
        if (
            vars_gene1 != {}
            and vars_gene2 != {}
            and list(vars_gene1.keys())[0] < list(vars_gene2.keys())[0]
        ):
            self.merge_vcf([(vars_gene1, gene1_hap_ids), (vars_gene2, gene2_hap_ids)])
        else:
            self.merge_vcf([(vars_gene2, gene2_hap_ids), (vars_gene1, gene1_hap_ids)])



# paraphase
# Author: Xiao Chen <xchen@pacificbiosciences.com>


from collections import namedtuple
import copy
import json
from ..phaser import Phaser


class Smn1Phaser(Phaser):
    fields = (
        "smn1_cn",
        "smn2_cn",
        "smn2_del78_cn",
        "smn1_read_number",
        "smn2_read_number",
        "smn2_del78_read_number",
        "highest_total_cn",
        "smn1_haplotypes",
        "smn2_haplotypes",
        "smn2_del78_haplotypes",
        "assembled_haplotypes",
        "two_copy_haplotypes",
        "sites_for_phasing",
        "unique_supporting_reads",
        "het_sites_not_used_in_phasing",
        "homozygous_sites",
        "haplotype_details",
        "nonunique_supporting_reads",
        "read_details",
        "final_haplotypes",
        "genome_depth",
        "region_depth",
    )
    GeneCall = namedtuple(
        "GeneCall",
        fields,
        defaults=(None,) * len(fields),
    )
    HaplotypeInfo = namedtuple(
        "HaplotypeInfo", "variants boundary boundary_gene2 haplogroup is_truncated"
    )

    def __init__(
        self, sample_id, outdir, genome_depth=None, genome_bam=None, sample_sex=None
    ):
        Phaser.__init__(self, sample_id, outdir, genome_depth, genome_bam, sample_sex)
        self.has_smn1 = False
        self.has_smn2 = False
        self.smn1_reads = set()
        self.smn2_reads = set()
        self.smn1_reads_splice = None
        self.smn2_reads_splice = None
        self.smn1_del_reads = set()
        self.smn1_del_reads_partial = set()
        self.smn2_del_reads = set()
        self.smn2_del_reads_partial = set()

    def set_parameter(self, config):
        super().set_parameter(config)
        self.deletion1_size = config["deletion1_size"]
        self.deletion2_size = config["deletion2_size"]
        self.deletion1_name = config["deletion1_name"]
        self.deletion2_name = config["deletion2_name"]
        self.del2_3p_pos1 = config["del2_3p_pos1"]
        self.del2_3p_pos2 = config["del2_3p_pos2"]
        self.del2_5p_pos1 = config["del2_5p_pos1"]
        self.del2_5p_pos2 = config["del2_5p_pos2"]
        self.del1_3p_pos1 = config["del1_3p_pos1"]
        self.del1_3p_pos2 = config["del1_3p_pos2"]
        self.del1_5p_pos1 = config["del1_5p_pos1"]
        self.del1_5p_pos2 = config["del1_5p_pos2"]
        self.strip_c_region_start = config["strip_c_region_start"]
        self.strip_c_region_end = config["strip_c_region_end"]
        with open(config["data"]["known_haplotypes"]) as f:
            self.known_haps = json.load(f)

    def check_smn1_smn2_presence(self):
        """
        Get number of reads at splice variant sites and
        check if there is any SMN1/SMN2
        """
        bamh = self._bamh
        pivot_site = self.pivot_site
        for pileupcolumn in bamh.pileup(
            self.nchr, pivot_site - 1, pivot_site, truncate=True
        ):
            for pileupread in pileupcolumn.pileups:
                if not pileupread.is_del and not pileupread.is_refskip:
                    read_seq = pileupread.alignment.query_sequence
                    read_pos = pileupread.query_position
                    base1 = read_seq[read_pos].upper()
                    if base1 == "C":
                        self.smn1_reads.add(pileupread.alignment.query_name)
                    elif base1 == "T" and read_seq[read_pos : read_pos + 2] != "TC":
                        self.smn2_reads.add(pileupread.alignment.query_name)

        self.smn1_reads_splice = len(self.smn1_reads)
        self.smn2_reads_splice = len(self.smn2_reads)
        if self.smn1_reads_splice >= 2:
            self.has_smn1 = True
        if len(self.smn2_del_reads_partial) + self.smn2_reads_splice >= 2:
            self.has_smn2 = True

    def allow_del_bases(self, pos):
        """
        During variant calling, allow calling bases at positions within big
        deletions
        """
        if (
            self.smn2_del_reads_partial != set()
            and self.del2_3p_pos1 <= pos <= self.del2_5p_pos2
        ):
            return True
        if (
            self.smn1_del_reads_partial != set()
            and self.del1_3p_pos1 <= pos <= self.del1_5p_pos2
        ):
            return True
        return False

    def output_variants_in_haplotypes(
        self, smn1_haps, smn2_haps, smn2_del_haps, reads, nonunique, two_cp_haps
    ):
        """
        Summarize all variants in each haplotype.
        Output all variants and their genotypes.
        Need to consider that haplotypes are different length
        """
        het_sites = self.het_sites
        haplotype_info = {}
        haplotype_variants = {}
        var_no_phasing = copy.deepcopy(self.het_no_phasing)
        for smn_haps in [smn1_haps, smn2_haps, smn2_del_haps]:
            for hap, hap_name in smn_haps.items():
                haplotype_variants.setdefault(hap_name, [])
        # het sites not used in phasing
        if reads != {}:
            for var in var_no_phasing:
                genotypes = []
                var_reads = self.check_variants_in_haplotypes(var)
                haps_with_variant = []
                for smn_haps in [smn1_haps, smn2_haps, smn2_del_haps]:
                    for hap, hap_name in smn_haps.items():
                        hap_reads = reads[hap]
                        hap_reads_nonunique = [
                            a for a in nonunique if hap in nonunique[a]
                        ]
                        genotype = self.get_genotype_in_hap(
                            var_reads, hap_reads, hap_reads_nonunique
                        )
                        genotypes.append(genotype)
                        if hap_name in two_cp_haps:
                            genotypes.append(genotype)
                        if genotype == "1":
                            haps_with_variant.append(hap_name)
                if haps_with_variant == []:
                    self.het_no_phasing.remove(var)
                else:
                    for hap_name in haps_with_variant:
                        haplotype_variants[hap_name].append(var)
        # het sites and homo sites
        for hap, hap_name in smn1_haps.items():
            for i in range(len(hap)):
                if hap[i] == "2":
                    haplotype_variants[hap_name].append(het_sites[i])
                elif (
                    hap[i] == "4"
                    and self.deletion1_name not in haplotype_variants[hap_name]
                ):
                    haplotype_variants[hap_name].append(self.deletion1_name)
            if self.deletion1_name in haplotype_variants[hap_name]:
                pos1 = self.del1_3p_pos1
                pos2 = self.del1_5p_pos2
                for var in self.homo_sites:
                    pos = int(var.split("_")[0])
                    if pos < pos1 or pos > pos2:
                        haplotype_variants[hap_name].append(var)
            else:
                haplotype_variants[hap_name] += self.homo_sites

            var_nstart, var_nend = self.get_hap_variant_ranges(hap)
            var_tmp = haplotype_variants[hap_name]
            var_tmp1 = [
                a for a in var_tmp if var_nstart <= int(a.split("_")[0]) <= var_nend
            ]
            var_tmp1 = list(set(var_tmp1))
            haplotype_variants[hap_name] = sorted(
                var_tmp1,
                key=lambda x: int(x.split("_")[0]),
            )
            haplotype_variants[hap_name].append((var_nstart, var_nend))
        for hap, hap_name in smn2_haps.items():
            for i in range(len(hap)):
                if hap[i] == "2":
                    haplotype_variants[hap_name].append(het_sites[i])
            haplotype_variants[hap_name] += self.homo_sites

            var_nstart, var_nend = self.get_hap_variant_ranges(hap)
            var_tmp = haplotype_variants[hap_name]
            var_tmp1 = [
                a for a in var_tmp if var_nstart <= int(a.split("_")[0]) <= var_nend
            ]
            var_tmp1 = list(set(var_tmp1))
            haplotype_variants[hap_name] = sorted(
                var_tmp1,
                key=lambda x: int(x.split("_")[0]),
            )
            haplotype_variants[hap_name].append((var_nstart, var_nend))
        for hap, hap_name in smn2_del_haps.items():
            for i in range(len(hap)):
                if hap[i] == "2":
                    haplotype_variants[hap_name].append(het_sites[i])
                elif (
                    hap[i] == "3"
                    and self.deletion2_name not in haplotype_variants[hap_name]
                ):
                    haplotype_variants[hap_name].append(self.deletion2_name)
            pos1 = self.del2_3p_pos1
            pos2 = self.del2_5p_pos2
            for var in self.homo_sites:
                pos = int(var.split("_")[0])
                if pos < pos1 or pos > pos2:
                    haplotype_variants[hap_name].append(var)

            var_nstart, var_nend = self.get_hap_variant_ranges(hap)
            var_tmp = haplotype_variants[hap_name]
            var_tmp1 = [
                a for a in var_tmp if var_nstart <= int(a.split("_")[0]) <= var_nend
            ]
            var_tmp1 = list(set(var_tmp1))
            haplotype_variants[hap_name] = sorted(
                var_tmp1,
                key=lambda x: int(x.split("_")[0]),
            )
            haplotype_variants[hap_name].append((var_nstart, var_nend))

        # assign haplogroups
        for hap_name in haplotype_variants:
            haplogroup = None
            if "del" not in hap_name:
                query_hap = haplotype_variants[hap_name]
                gene = hap_name[:4]
                haplogroup, candidates1, candidates2 = self.assign_hap_to_group(
                    query_hap, gene
                )
            else:
                haplogroup = "smn2_del_exon78"
            haplotype_info.setdefault(
                hap_name,
                self.HaplotypeInfo(
                    haplotype_variants[hap_name][:-1],
                    haplotype_variants[hap_name][-1],
                    [
                        self.get_range_in_other_gene(
                            haplotype_variants[hap_name][-1][0]
                        ),
                        self.get_range_in_other_gene(
                            haplotype_variants[hap_name][-1][1]
                        ),
                    ],
                    haplogroup,
                    False,
                )._asdict(),
            )

        return haplotype_info

    def assign_haps_to_gene(self, ass_haps):
        """Assign assembled haplotypes to smn1/smn2/smn2del"""
        splice_index, found_splice = self.get_pivot_site_index()
        smn1_haps = []
        smn2_haps = []
        smn2_del_haps = []

        for hap in ass_haps:
            if "3" in hap:
                smn2_del_haps.append(hap)

        if found_splice is True:
            for a in ass_haps:
                if a[splice_index] == "1":
                    smn1_haps.append(a)
                elif "3" not in a:
                    smn2_haps.append(a)
        elif self.has_smn1 is False or self.smn2_reads_splice < 2:
            for a in ass_haps:
                if "3" not in a:
                    if self.has_smn1 is True:
                        smn1_haps.append(a)
                    else:
                        smn2_haps.append(a)
        return smn1_haps, smn2_haps, smn2_del_haps

    def adjust_smn1_cn(self, smn1_cn, smn2_cn, hcn, ass_haps, read_counts, smn1_haps):
        """
        Adjust SMN1 CN if there is only one haplotype found.
        """
        new_smn1_cn = smn1_cn
        two_cp_haps = []
        if smn1_cn is None:
            return None, two_cp_haps
        if self.mdepth is not None:
            if self.mdepth < 20 and smn1_cn == 1:
                return None, two_cp_haps
            genome_depth = self.mdepth
            haploid_depth = genome_depth / 2
            depth1 = self.smn1_reads_splice
            copy_number_probs = self.depth_prob(depth1, haploid_depth)
            if smn1_cn == 1:
                # when there is only one haplotype, check if depth is
                # consistent with haploid depth
                copy_one_prob = copy_number_probs[0]
                if copy_one_prob < 0.05:
                    two_cp_haps = list(smn1_haps.values())
                    return 2, two_cp_haps
                if copy_one_prob < 0.25:
                    new_smn1_cn = None
            elif smn1_cn == 2:
                # scenario where two two-copy alleles are identical
                copy_four_prob = copy_number_probs[3]
                if copy_four_prob > 0.75:
                    two_cp_haps = list(smn1_haps.values())
                    return 4, two_cp_haps
            if smn1_cn in [2, 3] and read_counts is not None:
                # check if one smn1 haplotype has more reads than others
                haps = list(read_counts.keys())
                counts = list(read_counts.values())
                max_count = max(counts)
                cp2_hap = haps[counts.index(max_count)]
                if cp2_hap in smn1_haps:
                    others_max = sorted(counts, reverse=True)[1]
                    probs = self.depth_prob(max_count, others_max)
                    if probs[0] < 0.0005 and others_max >= 10:
                        two_cp_haps.append(smn1_haps[cp2_hap])
                        return smn1_cn + 1, two_cp_haps

        if smn1_cn == 1 and smn2_cn >= 1:
            haploid_depth = self.smn2_reads_splice / smn2_cn
            if haploid_depth >= 10:
                depth1 = self.smn1_reads_splice
                copy_number_probs = self.depth_prob(depth1, haploid_depth)
                # when there is only one haplotype, check if depth is
                # consistent with haploid depth
                copy_one_prob = copy_number_probs[0]
                if copy_one_prob < 0.25:
                    new_smn1_cn = None
                    if smn2_cn > 1 and copy_one_prob < 0.05:
                        two_cp_haps = list(smn1_haps.values())
                        return 2, two_cp_haps
        # if we see more haplotypes in other regions of the gene
        if smn1_cn == 1 and hcn > len(ass_haps):
            if self.has_smn2 is False:
                two_cp_haps = list(smn1_haps.values())
                return 2, two_cp_haps
            else:
                new_smn1_cn = None
        return new_smn1_cn, two_cp_haps

    def adjust_smn2_cn(self, smn1_cn, smn2_cn, smn2_haps):
        """
        Adjust SMN2 CN if there is only one haplotype found.
        """
        new_smn2_cn = smn2_cn
        two_cp_haps = []
        if smn2_cn is None:
            return None, two_cp_haps
        if self.mdepth is not None:
            if self.mdepth < 20 and smn2_cn == 1:
                return None, two_cp_haps
            genome_depth = self.mdepth
            haploid_depth = genome_depth / 2
            depth1 = self.smn2_reads_splice
            copy_number_probs = self.depth_prob(depth1, haploid_depth)
            # if smn1 cn is 3, then no need to adjust smn2 cn
            if (
                smn2_cn == 1
                and len(self.smn2_del_reads_partial) <= 1
                and (smn1_cn is None or smn1_cn <= 2)
            ):
                # when there is only one haplotype, check if depth is
                # consistent with haploid depth
                copy_one_prob = copy_number_probs[0]
                if copy_one_prob < 0.25:
                    new_smn2_cn = None
                    if copy_one_prob < 0.05:
                        two_cp_haps = list(smn2_haps.values())
                        return 2, two_cp_haps
        if smn1_cn is not None:
            if smn2_cn == 1 and smn1_cn == 2 and len(self.smn2_del_reads_partial) <= 1:
                haploid_depth = self.smn1_reads_splice / smn1_cn
                if haploid_depth > 10:
                    depth1 = self.smn2_reads_splice
                    copy_number_probs = self.depth_prob(depth1, haploid_depth)
                    # when there is only one haplotype, check if depth is
                    # consistent with haploid depth
                    copy_one_prob = copy_number_probs[0]
                    if copy_one_prob < 0.25:
                        new_smn2_cn = None
                        if copy_one_prob < 0.05:
                            two_cp_haps = list(smn2_haps.values())
                            return 2, two_cp_haps
        return new_smn2_cn, two_cp_haps

    def get_best_match(
        self,
        known_haps,
        query_hap,
        min_overlap_len=12000,
        strip_c=False,
        max_mismatch=15,
    ):
        """
        Get best match to a haplogroup
        """
        vars_all = query_hap[:-1]
        nstart, nend = query_hap[-1]
        haplogroup_candidates = {}
        for known_hap_read in known_haps:
            known_hap_vars_all = known_haps[known_hap_read][:-2]
            known_hap_nstart, known_hap_nend = known_haps[known_hap_read][-2]
            new_nstart = max(nstart, known_hap_nstart)
            new_nend = min(nend, known_hap_nend)
            if strip_c:
                new_nstart = max(new_nstart, self.strip_c_region_start)
                new_nend = min(new_nend, self.strip_c_region_end)
            if new_nend - new_nstart >= min_overlap_len:
                haplogroup = known_haps[known_hap_read][-1]
                if strip_c:
                    haplogroup = haplogroup.strip("c")
                var1 = set(
                    [
                        a
                        for a in vars_all
                        if new_nstart <= int(a.split("_")[0]) <= new_nend
                    ]
                )
                var2 = set(
                    [
                        a
                        for a in known_hap_vars_all
                        if new_nstart <= int(a.split("_")[0]) <= new_nend
                    ]
                )
                nmismatch = len(var1 - var2) + len(var2 - var1)
                haplogroup_candidates.setdefault(haplogroup, []).append(nmismatch)
        if haplogroup_candidates == {}:
            return None, {}
        tmp = {k: sorted(v) for k, v in haplogroup_candidates.items()}
        haplogroup_candidates = dict(sorted(tmp.items(), key=lambda item: min(item[1])))
        best_two_mismatches = list(haplogroup_candidates.values())[:2]
        best_match = None
        if (
            len(best_two_mismatches[0]) >= 2
            and (
                best_two_mismatches[0][1] <= max_mismatch
                or best_two_mismatches[0][0] <= 2
            )
            and best_two_mismatches[0][0] < best_two_mismatches[1][0] - 2
            and best_two_mismatches[0][1] < best_two_mismatches[1][0]
        ):
            best_match = list(haplogroup_candidates.keys())[0]
        return best_match, haplogroup_candidates

    def assign_hap_to_group(self, query_hap, gene):
        """
        Assign haplotype to a haplogroup
        """
        known_haps = self.known_haps[gene]
        best_match1, candidates1 = self.get_best_match(known_haps, query_hap)
        best_match2, candidates2 = self.get_best_match(
            known_haps, query_hap, strip_c=True, max_mismatch=10
        )
        if best_match1 is not None:
            best_match = best_match1
        elif best_match2 is not None:
            best_match = best_match2
        else:
            best_match = None
        if best_match is not None and best_match == "S1-9":
            if self.deletion1_name in query_hap:
                best_match = "S1-9d"
        return best_match, candidates1, candidates2

    def call(self):
        """
        Main function that calls SMN1/SMN2 copy number and variants
        """
        if self.check_coverage_before_analysis() is False:
            return self.GeneCall()
        self.get_homopolymer()
        # find known deletions
        self.smn2_del_reads, self.smn2_del_reads_partial = self.get_long_del_reads(
            self.del2_3p_pos1,
            self.del2_3p_pos2,
            self.del2_5p_pos1,
            self.del2_5p_pos2,
            self.deletion2_size,
        )
        self.smn1_del_reads, self.smn1_del_reads_partial = self.get_long_del_reads(
            self.del1_3p_pos1,
            self.del1_3p_pos2,
            self.del1_5p_pos1,
            self.del1_5p_pos2,
            self.deletion1_size,
        )
        self.check_smn1_smn2_presence()
        if self.has_smn1 is False and self.has_smn2 is False:
            raise Exception("Not enough reads to analyze this region.")

        # scan for polymorphic sites
        regions_to_check = []
        if self.smn2_del_reads_partial != set():
            regions_to_check += [
                [self.del2_3p_pos1, self.del2_3p_pos2],
                [self.del2_5p_pos1, self.del2_5p_pos2],
            ]
        if self.smn1_del_reads_partial != set():
            regions_to_check += [
                [self.del1_3p_pos1, self.del1_3p_pos2],
                [self.del1_5p_pos1, self.del1_5p_pos2],
            ]
        self.get_candidate_pos(regions_to_check=regions_to_check)

        self.het_sites = sorted(list(self.candidate_pos))
        self.remove_noisy_sites()
        homo_sites_to_add = self.add_homo_sites()
        raw_read_haps = self.get_haplotypes_from_reads(
            kept_sites=homo_sites_to_add, add_sites=self.add_sites
        )

        # update reads for those overlapping known deletions
        het_sites = self.het_sites
        if len(self.smn2_del_reads_partial) > 1:
            raw_read_haps, het_sites = self.update_reads_for_deletions(
                raw_read_haps,
                het_sites,
                self.del2_3p_pos1,
                self.del2_5p_pos2,
                self.smn2_del_reads_partial,
                "3",
                self.deletion2_name,
            )
        if len(self.smn1_del_reads_partial) > 1:
            raw_read_haps, het_sites = self.update_reads_for_deletions(
                raw_read_haps,
                het_sites,
                self.del1_3p_pos1,
                self.del1_5p_pos2,
                self.smn1_del_reads_partial,
                "4",
                self.deletion1_name,
            )
        self.het_sites = het_sites

        # assemble haplotypes
        (
            ass_haps,
            original_haps,
            hcn,
            uniquely_supporting_reads,
            nonuniquely_supporting_reads,
            raw_read_haps,
            read_counts,
        ) = self.phase_haps(raw_read_haps)

        # process and assign haplotypes
        smn1_cn = None
        smn2_cn = None
        smn2_del_cn = 0
        smn1_haps, smn2_haps, smn2_del_haps = self.assign_haps_to_gene(ass_haps)
        if ass_haps == [] and self.het_sites == []:
            if self.has_smn1 is True and self.has_smn2 is False:
                smn1_cn = 1
                smn2_cn = 0
            elif self.has_smn1 is False and self.has_smn2 is True:
                smn1_cn = 0
                smn2_cn = 1
        else:
            smn1_cn = len(smn1_haps)
            smn2_cn = len(smn2_haps)
            smn2_del_cn = len(smn2_del_haps)

        tmp = {}
        for i, hap in enumerate(smn1_haps):
            tmp.setdefault(hap, f"smn1hap{i+1}")
        smn1_haps = tmp
        tmp = {}
        for i, hap in enumerate(smn2_haps):
            tmp.setdefault(hap, f"smn2hap{i+1}")
        smn2_haps = tmp
        tmp = {}
        for i, hap in enumerate(smn2_del_haps):
            tmp.setdefault(hap, f"smn2del78hap{i+1}")
        smn2_del_haps = tmp

        # update cn based on depth
        smn1_cn_old = smn1_cn
        smn1_cn, two_cp_haps = self.adjust_smn1_cn(
            smn1_cn, smn2_cn, hcn, ass_haps, read_counts, smn1_haps
        )
        smn2_cn, two_cp_haps_smn2 = self.adjust_smn2_cn(smn1_cn_old, smn2_cn, smn2_haps)
        two_cp_haps += two_cp_haps_smn2

        # summarize variants
        haplotypes = None
        if self.het_sites != []:
            haplotypes = self.output_variants_in_haplotypes(
                smn1_haps,
                smn2_haps,
                smn2_del_haps,
                uniquely_supporting_reads,
                nonuniquely_supporting_reads,
                two_cp_haps,
            )

        self.close_handle()

        return self.GeneCall(
            smn1_cn,
            smn2_cn,
            smn2_del_cn,
            self.smn1_reads_splice,
            self.smn2_reads_splice,
            len(self.smn2_del_reads_partial),
            hcn,
            smn1_haps,
            smn2_haps,
            smn2_del_haps,
            original_haps,
            two_cp_haps,
            self.het_sites,
            uniquely_supporting_reads,
            self.het_no_phasing,
            self.homo_sites,
            haplotypes,
            nonuniquely_supporting_reads,
            raw_read_haps,
            {**smn1_haps, **smn2_haps, **smn2_del_haps},
            self.mdepth,
            self.region_avg_depth._asdict(),
        )
